"""
This module provides functions to calculate various seawater properties.

Functions:
    ctd2svel: Calculate sound velocity from conductivity, temperature, and depth.
    depth2pressure: Calculate pressure from depth.
    ctp2salinity: Calculate absolute salinity from conductivity, temperature, and pressure.
    ctd2salinity: Calculate absolute salinity from conductivity, temperature, and depth.
    ctp2pden: Calculate potential density from conductivity, temperature, and pressure.
    ctd2oxygen_saturation: Calculate oxygen saturation from conductivity, temperature, and depth.

Authors: Sebastián Rodríguez-Martínez and Giancarlo Troni
Contact: srodriguez@mbari.org
"""

from typing import List, Union

import gsw
import numpy as np


def ctd2svel(
    conductivity: Union[float, np.ndarray, List[float]],
    temperature: Union[float, np.ndarray, List[float]],
    depth: Union[float, np.ndarray, List[float]],
    latitude: Union[float, np.ndarray, List[float]] = None,
    longitude: Union[float, np.ndarray, List[float]] = None,
) -> Union[float, np.ndarray]:
    """
    Calculates sound velocity in m/s from conductivity, temperature, and depth.

    Args:
        conductivity (Union[float, np.ndarray, List[float]]): Conductivity in S/m.
        temperature (Union[float, np.ndarray, List[float]]): Temperature in degrees Celsius (ITS-90).
        depth (Union[float, np.ndarray, List[float]]): Depth in meters. Positive is upward.
        latitude (Union[float, np.ndarray, List[float]], optional): Latitude in degrees. Defaults to None.
        longitude (Union[float, np.ndarray, List[float]], optional): Longitude in degrees. Defaults to None.
    Returns:
        sound_velocity (Union[float, np.ndarray]): Sound velocity in m/s.

    Raises:
        TypeError: If the inputs are not of the correct type.
        ValueError: If the inputs are not of the correct shape or range.
    """
    # Convert to numpy arrays
    if isinstance(conductivity, list):
        conductivity = np.array(conductivity)
    if isinstance(temperature, list):
        temperature = np.array(temperature)
    if isinstance(depth, list):
        depth = np.array(depth)
    if isinstance(latitude, list):
        latitude = np.array(latitude)
    if isinstance(longitude, list):
        longitude = np.array(longitude)

    # Check inputs type
    if not isinstance(conductivity, (np.ndarray, float)):
        raise TypeError("Conductivity must be a float or an array.")
    if not isinstance(temperature, (np.ndarray, float)):
        raise TypeError("Temperature must be a float or an array.")
    if not isinstance(depth, (np.ndarray, float)):
        raise TypeError("Depth must be a float or an array.")
    if latitude is not None and not isinstance(latitude, (np.ndarray, float)):
        raise TypeError("Latitude must be a float or an array.")
    if longitude is not None and not isinstance(longitude, (np.ndarray, float)):
        raise TypeError("Longitude must be a float or an array.")

    # Check shape of inputs if they are numpy arrays
    if isinstance(conductivity, np.ndarray):
        conductivity = conductivity.squeeze()
        if conductivity.ndim > 1:
            raise ValueError("Conductivity must be a 1D array.")
    if isinstance(temperature, np.ndarray):
        temperature = temperature.squeeze()
        if temperature.ndim > 1:
            raise ValueError("Temperature must be a 1D array.")
    if isinstance(depth, np.ndarray):
        depth = depth.squeeze()
        if depth.ndim > 1:
            raise ValueError("Depth must be a 1D array.")
    if latitude is not None and isinstance(latitude, np.ndarray):
        latitude = latitude.squeeze()
        if latitude.ndim > 1:
            raise ValueError("Latitude must be a 1D array.")
    if longitude is not None and isinstance(longitude, np.ndarray):
        longitude = longitude.squeeze()
        if longitude.ndim > 1:
            raise ValueError("Longitude must be a 1D array.")

    # If latitude is not provided, set it to 0
    if latitude is None:
        latitude = 0 if isinstance(depth, float) else np.zeros_like(depth)
    elif isinstance(latitude, float) and isinstance(depth, np.ndarray):
        latitude = np.full_like(depth, latitude)

    # If longitude is not provided, set it to 0
    if longitude is None:
        longitude = 0 if isinstance(depth, float) else np.zeros_like(depth)
    elif isinstance(longitude, float) and isinstance(depth, np.ndarray):
        longitude = np.full_like(depth, longitude)

    # Check that the inputs are of the same shape
    if isinstance(conductivity, np.ndarray):
        if (
            conductivity.shape != temperature.shape
            or conductivity.shape != depth.shape
            or conductivity.shape != latitude.shape
            or conductivity.shape != longitude.shape
        ):
            raise ValueError(
                "Conductivity, temperature, depth, latitude, and longitude must have the same shape."
            )

    # Check that the depth is negative
    if isinstance(depth, np.ndarray):
        if np.any(depth > 0):
            raise ValueError("Depth must be negative.")
    else:
        if depth > 0:
            raise ValueError("Depth must be negative.")

    # Check that the latitude is between -90 and 90
    if isinstance(latitude, np.ndarray):
        if np.any(latitude < -90) or np.any(latitude > 90):
            raise ValueError("Latitude must be between -90 and 90 degrees.")
    else:
        if latitude < -90 or latitude > 90:
            raise ValueError("Latitude must be between -90 and 90 degrees.")

    # Check that the longitude is between -360 and 360
    if isinstance(longitude, np.ndarray):
        if np.any(longitude < -360) or np.any(longitude > 360):
            raise ValueError("Longitude must be between -360 and 360 degrees.")
    else:
        if longitude < -360 or longitude > 360:
            raise ValueError("Longitude must be between -360 and 360 degrees.")

    # Calculate sound velocity
    pressure = gsw.p_from_z(depth, latitude)
    practical_salinity = gsw.SP_from_C(conductivity, temperature, pressure)
    absolute_salinity = gsw.SA_from_SP(
        practical_salinity, pressure, longitude, latitude
    )
    conservative_temperature = gsw.CT_from_t(absolute_salinity, temperature, pressure)
    sound_speed = gsw.sound_speed(absolute_salinity, conservative_temperature, pressure)
    return sound_speed


def depth2pressure(
    depth: Union[float, np.ndarray, List[float]],
    latitude: Union[float, np.ndarray, List[float]] = None,
) -> Union[float, np.ndarray]:
    """
    Calculates pressure in dbars from depth in meters.

    Args:
        depth (Union[float, np.ndarray, List[float]]): Depth in meters.
        latitude (Union[float, np.ndarray, List[float]], optional): Latitude in degrees. Defaults to None.

    Returns:
        pressure (Union[float, np.ndarray]): Pressure in dbars.

    Raises:
        ValueError: If the latitude is not between -90 and 90 degrees.
    """
    # Convert to numpy arrays
    if isinstance(depth, list):
        depth = np.array(depth)
    if isinstance(latitude, list):
        latitude = np.array(latitude)

    # Check inputs type
    if not isinstance(depth, (np.ndarray, float)):
        raise TypeError("Depth must be a float or an array.")
    if latitude is not None and not isinstance(latitude, (np.ndarray, float)):
        raise TypeError("Latitude must be a float or an array.")

    # Check shape of inputs if they are numpy arrays
    if isinstance(depth, np.ndarray):
        depth = depth.squeeze()
        if depth.ndim > 1:
            raise ValueError("Depth must be a 1D array.")
    if latitude is not None and isinstance(latitude, np.ndarray):
        latitude = latitude.squeeze()
        if latitude.ndim > 1:
            raise ValueError("Latitude must be a 1D array.")

    # If latitude is not provided, set it to 0
    if latitude is None:
        latitude = 0 if isinstance(depth, float) else np.zeros_like(depth)
    elif isinstance(latitude, float) and isinstance(depth, np.ndarray):
        latitude = np.full_like(depth, latitude)

    # Check that the inputs are of the same shape
    if isinstance(depth, np.ndarray):
        if depth.shape != latitude.shape:
            raise ValueError("Depth and latitude must have the same shape.")

    # Check that the depth is negative
    if isinstance(depth, np.ndarray):
        if np.any(depth > 0):
            raise ValueError("Depth must be negative.")
    else:
        if depth > 0:
            raise ValueError("Depth must be negative.")

    # Check that the latitude is beetwen -90 and 90
    if isinstance(latitude, np.ndarray):
        if np.any(latitude < -90) or np.any(latitude > 90):
            raise ValueError("Latitude must be between -90 and 90 degrees.")
    else:
        if latitude < -90 or latitude > 90:
            raise ValueError("Latitude must be between -90 and 90 degrees.")

    return gsw.p_from_z(depth, latitude)


def ctp2salinity(
    conductivity: Union[float, np.ndarray, List[float]],
    temperature: Union[float, np.ndarray, List[float]],
    pressure: Union[float, np.ndarray, List[float]],
    latitude: Union[float, np.ndarray, List[float]] = None,
    longitude: Union[float, np.ndarray, List[float]] = None,
) -> Union[float, np.ndarray]:
    """
     Calculates absolute salinity in g/kg from conductivity, temperature, and pressure.

     Args:
        conductivity (Union[float, np.ndarray, List[float]]): Conductivity in S/m.
        temperature (Union[float, np.ndarray, List[float]]): Temperature in degrees Celsius (ITS-90).
        pressure (Union[float, np.ndarray, List[float]]): Pressure in dbars.
        latitude (Union[float, np.ndarray, List[float]], optional): Latitude in degrees. Defaults to None.
        longitude (Union[float, np.ndarray, List[float]], optional): Longitude in degrees. Defaults to None.

    Returns:
        absolute_salinity (Union[float, np.ndarray]): Absolute salinity in g/kg.

    Raises:
        TypeError: If the inputs are not of the correct type.
        ValueError: If the inputs are not of the correct shape or range.
    """
    # Convert to numpy arrays
    if isinstance(conductivity, list):
        conductivity = np.array(conductivity)
    if isinstance(temperature, list):
        temperature = np.array(temperature)
    if isinstance(pressure, list):
        pressure = np.array(pressure)
    if isinstance(latitude, list):
        latitude = np.array(latitude)
    if isinstance(longitude, list):
        longitude = np.array(longitude)

    # Check inputs type
    if not isinstance(conductivity, (np.ndarray, float)):
        raise TypeError("Conductivity must be a float or an array.")
    if not isinstance(temperature, (np.ndarray, float)):
        raise TypeError("Temperature must be a float or an array.")
    if not isinstance(pressure, (np.ndarray, float)):
        raise TypeError("Pressure must be a float or an array.")
    if latitude is not None and not isinstance(latitude, (np.ndarray, float)):
        raise TypeError("Latitude must be a float or an array.")
    if longitude is not None and not isinstance(longitude, (np.ndarray, float)):
        raise TypeError("Longitude must be a float or an array.")

    # Check shape of inputs if they are numpy arrays
    if isinstance(conductivity, np.ndarray):
        conductivity = conductivity.squeeze()
        if conductivity.ndim > 1:
            raise ValueError("Conductivity must be a 1D array.")
    if isinstance(temperature, np.ndarray):
        temperature = temperature.squeeze()
        if temperature.ndim > 1:
            raise ValueError("Temperature must be a 1D array.")
    if isinstance(pressure, np.ndarray):
        pressure = pressure.squeeze()
        if pressure.ndim > 1:
            raise ValueError("Pressure must be a 1D array.")
    if latitude is not None and isinstance(latitude, np.ndarray):
        latitude = latitude.squeeze()
        if latitude.ndim > 1:
            raise ValueError("Latitude must be a 1D array.")
    if longitude is not None and isinstance(longitude, np.ndarray):
        longitude = longitude.squeeze()
        if longitude.ndim > 1:
            raise ValueError("Longitude must be a 1D array.")

    # If latitude is not provided, set it to 0
    if latitude is None:
        latitude = 0 if isinstance(pressure, float) else np.zeros_like(pressure)
    elif isinstance(latitude, float) and isinstance(pressure, np.ndarray):
        latitude = np.full_like(pressure, latitude)

    # If longitude is not provided, set it to 0
    if longitude is None:
        longitude = 0 if isinstance(pressure, float) else np.zeros_like(pressure)
    elif isinstance(longitude, float) and isinstance(pressure, np.ndarray):
        longitude = np.full_like(pressure, longitude)

    # Check that the inputs are of the same shape
    if isinstance(conductivity, np.ndarray):
        if (
            conductivity.shape != temperature.shape
            or conductivity.shape != pressure.shape
            or conductivity.shape != latitude.shape
            or conductivity.shape != longitude.shape
        ):
            raise ValueError(
                "Conductivity, temperature, pressure, latitude, and longitude must have the same shape."
            )

    # Check that the latitude is beetwen -90 and 90
    if isinstance(latitude, np.ndarray):
        if np.any(latitude < -90) or np.any(latitude > 90):
            raise ValueError("Latitude must be between -90 and 90 degrees.")
    else:
        if latitude < -90 or latitude > 90:
            raise ValueError("Latitude must be between -90 and 90 degrees.")

    # Check that the longitude is beetwen -360 and 360
    if isinstance(longitude, np.ndarray):
        if np.any(longitude < -360) or np.any(longitude > 360):
            raise ValueError("Longitude must be between -360 and 360 degrees.")
    else:
        if longitude < -360 or longitude > 360:
            raise ValueError("Longitude must be between -360 and 360 degrees.")
    practical_salinity = gsw.SP_from_C(conductivity, temperature, pressure)
    absolute_salinity = gsw.SA_from_SP(
        practical_salinity, pressure, longitude, latitude
    )
    return absolute_salinity


def ctd2salinity(
    conductivity: Union[float, np.ndarray, List[float]],
    temperature: Union[float, np.ndarray, List[float]],
    depth: Union[float, np.ndarray, List[float]],
    latitude: Union[float, np.ndarray, List[float]] = None,
    longitude: Union[float, np.ndarray, List[float]] = None,
) -> Union[float, np.ndarray]:
    """
    Calculates absolute salinity in g/kg from conductivity, temperature, and depth.

    Args:
        conductivity (Union[float, np.ndarray, List[float]]): Conductivity in S/m.
        temperature (Union[float, np.ndarray, List[float]]): Temperature in degrees Celsius (ITS-90).
        depth (Union[float, np.ndarray, List[float]]): Depth in meters. Positive is upward.
        latitude (Union[float, np.ndarray, List[float]], optional): Latitude in degrees. Defaults to None.
        longitude (Union[float, np.ndarray, List[float]], optional): Longitude in degrees. Defaults to None.
    Returns:
        absolute_salinity (Union[float, np.ndarray]): Absolute salinity in g/kg.

    Raises:
        TypeError: If the inputs are not of the correct type.
        ValueError: If the inputs are not of the correct shape or range.
    """
    # Convert to numpy arrays
    if isinstance(conductivity, list):
        conductivity = np.array(conductivity)
    if isinstance(temperature, list):
        temperature = np.array(temperature)
    if isinstance(depth, list):
        depth = np.array(depth)
    if isinstance(latitude, list):
        latitude = np.array(latitude)
    if isinstance(longitude, list):
        longitude = np.array(longitude)

    # Check inputs type
    if not isinstance(conductivity, (np.ndarray, float)):
        raise TypeError("Conductivity must be a float or an array.")
    if not isinstance(temperature, (np.ndarray, float)):
        raise TypeError("Temperature must be a float or an array.")
    if not isinstance(depth, (np.ndarray, float)):
        raise TypeError("Depth must be a float or an array.")
    if latitude is not None and not isinstance(latitude, (np.ndarray, float)):
        raise TypeError("Latitude must be a float or an array.")
    if longitude is not None and not isinstance(longitude, (np.ndarray, float)):
        raise TypeError("Longitude must be a float or an array.")

    # Check shape of inputs if they are numpy arrays
    if isinstance(conductivity, np.ndarray):
        conductivity = conductivity.squeeze()
        if conductivity.ndim > 1:
            raise ValueError("Conductivity must be a 1D array.")
    if isinstance(temperature, np.ndarray):
        temperature = temperature.squeeze()
        if temperature.ndim > 1:
            raise ValueError("Temperature must be a 1D array.")
    if isinstance(depth, np.ndarray):
        depth = depth.squeeze()
        if depth.ndim > 1:
            raise ValueError("Depth must be a 1D array.")
    if latitude is not None and isinstance(latitude, np.ndarray):
        latitude = latitude.squeeze()
        if latitude.ndim > 1:
            raise ValueError("Latitude must be a 1D array.")
    if longitude is not None and isinstance(longitude, np.ndarray):
        longitude = longitude.squeeze()
        if longitude.ndim > 1:
            raise ValueError("Longitude must be a 1D array.")

    # If latitude is not provided, set it to 0
    if latitude is None:
        latitude = 0 if isinstance(depth, float) else np.zeros_like(depth)
    elif isinstance(latitude, float) and isinstance(depth, np.ndarray):
        latitude = np.full_like(depth, latitude)

    # If longitude is not provided, set it to 0
    if longitude is None:
        longitude = 0 if isinstance(depth, float) else np.zeros_like(depth)
    elif isinstance(longitude, float) and isinstance(depth, np.ndarray):
        longitude = np.full_like(depth, longitude)

    # Check that the inputs are of the same shape
    if isinstance(conductivity, np.ndarray):
        if (
            conductivity.shape != temperature.shape
            or conductivity.shape != depth.shape
            or conductivity.shape != latitude.shape
            or conductivity.shape != longitude.shape
        ):
            raise ValueError(
                "Conductivity, temperature, depth, latitude, and longitude must have the same shape."
            )

    # Check that the depth is negative
    if isinstance(depth, np.ndarray):
        if np.any(depth > 0):
            raise ValueError("Depth must be negative.")
    else:
        if depth > 0:
            raise ValueError("Depth must be negative.")

    # Check that the latitude is beetwen -90 and 90
    if isinstance(latitude, np.ndarray):
        if np.any(latitude < -90) or np.any(latitude > 90):
            raise ValueError("Latitude must be between -90 and 90 degrees.")
    else:
        if latitude < -90 or latitude > 90:
            raise ValueError("Latitude must be between -90 and 90 degrees.")

    # Check that the longitude is beetwen -360 and 360
    if isinstance(longitude, np.ndarray):
        if np.any(longitude < -360) or np.any(longitude > 360):
            raise ValueError("Longitude must be between -360 and 360 degrees.")
    else:
        if longitude < -360 or longitude > 360:
            raise ValueError("Longitude must be between -360 and 360 degrees.")

    # Calculate sound velocity
    pressure = gsw.p_from_z(depth, latitude)
    practical_salinity = gsw.SP_from_C(conductivity, temperature, pressure)
    absolute_salinity = gsw.SA_from_SP(
        practical_salinity, pressure, longitude, latitude
    )
    return absolute_salinity


def ctp2pden(
    conductivity: Union[float, np.ndarray, List[float]],
    temperature: Union[float, np.ndarray, List[float]],
    pressure: Union[float, np.ndarray, List[float]],
    latitude: Union[float, np.ndarray, List[float]] = None,
    longitude: Union[float, np.ndarray, List[float]] = None,
    pressure_reference: Union[float, np.ndarray, List[float]] = None,
) -> Union[float, np.ndarray]:
    """
    Calculates potential density (kg/m^3) relative to a reference pressure
    (default 0 dbar) from conductivity, temperature, and pressure using the gsw library.

    Args:
        conductivity (Union[float, np.ndarray, List[float]]): Conductivity in S/m.
        temperature (Union[float, np.ndarray, List[float]]): Temperature in degrees Celsius (ITS-90).
        pressure (Union[float, np.ndarray, List[float]]): Pressure in dbars.
        latitude (Union[float, np.ndarray, List[float]], optional): Latitude in degrees. Defaults to None.
        longitude (Union[float, np.ndarray, List[float]], optional): Longitude in degrees. Defaults to None.
        pressure_reference (Union[float, np.ndarray, List[float]], optional): Reference pressure in dbars. Defaults to None.

    Returns:
        potential_density (Union[float, np.ndarray]): Potential density in kg/m^3.

    Raises:
        TypeError: If the inputs are not of the correct type.
        ValueError: If the inputs are not of the correct shape or range.
    """
    # Convert to numpy arrays
    if isinstance(conductivity, list):
        conductivity = np.array(conductivity)
    if isinstance(temperature, list):
        temperature = np.array(temperature)
    if isinstance(pressure, list):
        pressure = np.array(pressure)
    if isinstance(latitude, list):
        latitude = np.array(latitude)
    if isinstance(longitude, list):
        longitude = np.array(longitude)
    if isinstance(pressure_reference, list):
        pressure_reference = np.array(pressure_reference)

    # Check inputs type
    if not isinstance(conductivity, (np.ndarray, float)):
        raise TypeError("Conductivity must be a float or an array.")
    if not isinstance(temperature, (np.ndarray, float)):
        raise TypeError("Temperature must be a float or an array.")
    if not isinstance(pressure, (np.ndarray, float)):
        raise TypeError("Pressure must be a float or an array.")
    if latitude is not None and not isinstance(latitude, (np.ndarray, float)):
        raise TypeError("Latitude must be a float or an array.")
    if longitude is not None and not isinstance(longitude, (np.ndarray, float)):
        raise TypeError("Longitude must be a float or an array.")
    if pressure_reference is not None and not isinstance(
        pressure_reference, (np.ndarray, float)
    ):
        raise TypeError("Pressure reference must be a float or an array.")

    # Check shape of inputs if they are numpy arrays
    if isinstance(conductivity, np.ndarray):
        conductivity = conductivity.squeeze()
        if conductivity.ndim > 1:
            raise ValueError("Conductivity must be a 1D array.")
    if isinstance(temperature, np.ndarray):
        temperature = temperature.squeeze()
        if temperature.ndim > 1:
            raise ValueError("Temperature must be a 1D array.")
    if isinstance(pressure, np.ndarray):
        pressure = pressure.squeeze()
        if pressure.ndim > 1:
            raise ValueError("Pressure must be a 1D array.")
    if latitude is not None and isinstance(latitude, np.ndarray):
        latitude = latitude.squeeze()
        if latitude.ndim > 1:
            raise ValueError("Latitude must be a 1D array.")
    if longitude is not None and isinstance(longitude, np.ndarray):
        longitude = longitude.squeeze()
        if longitude.ndim > 1:
            raise ValueError("Longitude must be a 1D array.")
    if pressure_reference is not None and isinstance(pressure_reference, np.ndarray):
        pressure_reference = pressure_reference.squeeze()
        if pressure_reference.ndim > 1:
            raise ValueError("Pressure reference must be a 1D array.")

    # If latitude is not provided, set it to 0
    if latitude is None:
        latitude = 0 if isinstance(pressure, float) else np.zeros_like(pressure)
    elif isinstance(latitude, float) and isinstance(pressure, np.ndarray):
        latitude = np.full_like(pressure, latitude)

    # If longitude is not provided, set it to 0
    if longitude is None:
        longitude = 0 if isinstance(pressure, float) else np.zeros_like(pressure)
    elif isinstance(longitude, float) and isinstance(pressure, np.ndarray):
        longitude = np.full_like(pressure, longitude)

    # If pressure reference is not provided, set it to 0
    if pressure_reference is None:
        pressure_reference = (
            0 if isinstance(pressure, float) else np.zeros_like(pressure)
        )
    elif isinstance(pressure_reference, float) and isinstance(pressure, np.ndarray):
        pressure_reference = np.full_like(pressure, pressure_reference)

    # Check that the inputs are of the same shape
    if isinstance(conductivity, np.ndarray):
        if (
            conductivity.shape != temperature.shape
            or conductivity.shape != pressure.shape
            or conductivity.shape != latitude.shape
            or conductivity.shape != longitude.shape
            or conductivity.shape != pressure_reference.shape
        ):
            raise ValueError(
                "Conductivity, temperature, pressure, latitude, longitude and reference pressure must have the same shape."
            )

    # Check that the latitude is beetwen -90 and 90
    if isinstance(latitude, np.ndarray):
        if np.any(latitude < -90) or np.any(latitude > 90):
            raise ValueError("Latitude must be between -90 and 90 degrees.")
    else:
        if latitude < -90 or latitude > 90:
            raise ValueError("Latitude must be between -90 and 90 degrees.")

    # Check that the longitude is beetwen -360 and 360
    if isinstance(longitude, np.ndarray):
        if np.any(longitude < -360) or np.any(longitude > 360):
            raise ValueError("Longitude must be between -360 and 360 degrees.")
    else:
        if longitude < -360 or longitude > 360:
            raise ValueError("Longitude must be between -360 and 360 degrees.")

    # Compute potential density using the exact potential density function
    practical_salinity = gsw.SP_from_C(conductivity, temperature, pressure)
    absolute_salinity = gsw.SA_from_SP(
        practical_salinity, pressure, longitude, latitude
    )
    pden = gsw.pot_rho_t_exact(
        absolute_salinity, temperature, pressure, pressure_reference
    )
    return pden


def ctd2oxygen_saturation(
    conductivity: Union[float, np.ndarray, List[float]],
    temperature: Union[float, np.ndarray, List[float]],
    depth: Union[float, np.ndarray, List[float]],
    latitude: Union[float, np.ndarray, List[float]] = None,
    longitude: Union[float, np.ndarray, List[float]] = None,
) -> Union[float, np.ndarray]:
    """
    Calculates the oxygen saturation value (ml/l) from salinity and temperature.

    Calculates the oxygen concentration expected at equilibrium with air at an
    Absolute Pressure of 101325 Pa (sea pressure of 0 dbar) including saturated
    water vapor. This function uses the solubility coefficients derived from the
    data of Benson and Krause (1984), as fitted by Garcia and Gordon (1992, 1993).

    Args:
        conductivity (Union[float, np.ndarray, List[float]]): Conductivity in S/m.
        temperature (Union[float, np.ndarray, List[float]]): Temperature in degrees Celsius (ITS-90).
        depth (Union[float, np.ndarray, List[float]]): Depth in meters. Positive is upward.
        latitude (Union[float, np.ndarray, List[float]], optional): Latitude in degrees. Defaults to None.
        longitude (Union[float, np.ndarray, List[float]], optional): Longitude in degrees. Defaults to None.

    Returns:
        solubility (Union[float, np.ndarray]): Solubility of oxygen in micro-moles per kg

    Raises:
        TypeError: If the inputs are not of the correct type.
        ValueError: If the inputs are not of the correct shape or range.
    """
    # Convert to numpy arrays
    if isinstance(conductivity, list):
        conductivity = np.array(conductivity)
    if isinstance(temperature, list):
        temperature = np.array(temperature)
    if isinstance(depth, list):
        depth = np.array(depth)
    if isinstance(latitude, list):
        latitude = np.array(latitude)
    if isinstance(longitude, list):
        longitude = np.array(longitude)

    # Check inputs type
    if not isinstance(conductivity, (np.ndarray, float)):
        raise TypeError("Conductivity must be a float or an array.")
    if not isinstance(temperature, (np.ndarray, float)):
        raise TypeError("Temperature must be a float or an array.")
    if not isinstance(depth, (np.ndarray, float)):
        raise TypeError("Depth must be a float or an array.")
    if latitude is not None and not isinstance(latitude, (np.ndarray, float)):
        raise TypeError("Latitude must be a float or an array.")
    if longitude is not None and not isinstance(longitude, (np.ndarray, float)):
        raise TypeError("Longitude must be a float or an array.")

    # Check shape of inputs if they are numpy arrays
    if isinstance(conductivity, np.ndarray):
        conductivity = conductivity.squeeze()
        if conductivity.ndim > 1:
            raise ValueError("Conductivity must be a 1D array.")
    if isinstance(temperature, np.ndarray):
        temperature = temperature.squeeze()
        if temperature.ndim > 1:
            raise ValueError("Temperature must be a 1D array.")
    if isinstance(depth, np.ndarray):
        depth = depth.squeeze()
        if depth.ndim > 1:
            raise ValueError("Depth must be a 1D array.")
    if latitude is not None and isinstance(latitude, np.ndarray):
        latitude = latitude.squeeze()
        if latitude.ndim > 1:
            raise ValueError("Latitude must be a 1D array.")
    if longitude is not None and isinstance(longitude, np.ndarray):
        longitude = longitude.squeeze()
        if longitude.ndim > 1:
            raise ValueError("Longitude must be a 1D array.")

    # If latitude is not provided, set it to 0
    if latitude is None:
        latitude = 0 if isinstance(depth, float) else np.zeros_like(depth)
    elif isinstance(latitude, float) and isinstance(depth, np.ndarray):
        latitude = np.full_like(depth, latitude)

    # If longitude is not provided, set it to 0
    if longitude is None:
        longitude = 0 if isinstance(depth, float) else np.zeros_like(depth)
    elif isinstance(longitude, float) and isinstance(depth, np.ndarray):
        longitude = np.full_like(depth, longitude)

    # Check that the inputs are of the same shape
    if isinstance(conductivity, np.ndarray):
        if (
            conductivity.shape != temperature.shape
            or conductivity.shape != depth.shape
            or conductivity.shape != latitude.shape
            or conductivity.shape != longitude.shape
        ):
            raise ValueError(
                "Conductivity, temperature, depth, latitude, and longitude must have the same shape."
            )

    # Check that the depth is negative
    if isinstance(depth, np.ndarray):
        if np.any(depth > 0):
            raise ValueError("Depth must be negative.")
    else:
        if depth > 0:
            raise ValueError("Depth must be negative.")

    # Check that the latitude is between -90 and 90
    if isinstance(latitude, np.ndarray):
        if np.any(latitude < -90) or np.any(latitude > 90):
            raise ValueError("Latitude must be between -90 and 90 degrees.")
    else:
        if latitude < -90 or latitude > 90:
            raise ValueError("Latitude must be between -90 and 90 degrees.")

    # Check that the longitude is between -360 and 360
    if isinstance(longitude, np.ndarray):
        if np.any(longitude < -360) or np.any(longitude > 360):
            raise ValueError("Longitude must be between -360 and 360 degrees.")
    else:
        if longitude < -360 or longitude > 360:
            raise ValueError("Longitude must be between -360 and 360 degrees.")

    # Calculate sound velocity
    pressure = gsw.p_from_z(depth, latitude)
    practical_salinity = gsw.SP_from_C(conductivity, temperature, pressure)
    absolute_salinity = gsw.SA_from_SP(
        practical_salinity, pressure, longitude, latitude
    )
    conservative_temperature = gsw.CT_from_t(absolute_salinity, temperature, pressure)
    o2sol = gsw.O2sol(
        absolute_salinity, conservative_temperature, pressure, longitude, latitude
    )
    return o2sol



# import methods from the corresponding modules
from .geography import (
    deg2m_lat,
    deg2m_lon,
    dms_to_decimal,
    get_local_gravity,
    latlon_to_zone_letter,
    latlon_to_zone_number,
    ll2utm,
    ll2xy,
    utm2ll,
    xy2ll,
)

# Get __all__ from the corresponding modules
__all__ = [
    "latlon_to_zone_number",
    "latlon_to_zone_letter",
    "ll2utm",
    "utm2ll",
    "deg2m_lat",
    "deg2m_lon",
    "ll2xy",
    "get_local_gravity",
    "xy2ll",
    "dms_to_decimal",
]



#!/usr/bin/python

"""
This script is used to scan the python path for LCM types. It searches for python files that contain the string
"_get_packed_fingerprint" and imports the modules that contain the string. The script then checks if the module
contains a class with the right name and methods. If the module passes the checks, it is added to the list of LCM types.
This script is based on lcm-log2smat is based on libbot2 script bot-log2mat.

Authors: Sebastián Rodríguez-Martínez, Giancarlo Troni and Bastián Muñoz
Contact: srodriguez@mbari.org
Last Update: 2024-03-25
"""

import os
import pyclbr
import re
import sys
from io import open
from typing import Any, Dict, List, Union


def find_lcmtypes(lcm_packages: List[str] = []) -> List[str]:
    """
    Walks the directory tree looking for python files that contain the string "_get_packed_fingerprint"
    and imports the modules that contain the string. The function then checks if the module contains a class
    with the right name and methods. If the module passes the checks, it is added to the list of LCM types.

    Args:
        lcm_packages (List[str]): List of directories to search for LCM types in addition to the directories in the
        python path.

    Returns:
        lcmtypes (List[str]): List of LCM types
    """
    lcmtypes = []
    regex = re.compile("_get_packed_fingerprint")

    # Search for lcm types in all directories in the python path
    dirs_to_check = sys.path

    # Extend the list of directories to check with the specified lcm_packages
    for package in lcm_packages:
        if os.path.exists(package):
            dirs_to_check.append(package)

    for dir_name in dirs_to_check:
        # For each directory in the python path, walk the directory tree, looking
        # for python files that contain the string "_get_packed_fingerprint"
        for root, dirs, files in os.walk(dir_name):
            subdirs = root[len(dir_name) :].split(os.sep)
            subdirs = [s for s in subdirs if s]

            python_package = ".".join(subdirs)

            for fname in files:
                # Checks if the file is a python file (*.py)
                if not fname.endswith(".py"):
                    continue

                # Check if the file name is a valid python module name
                # 1. Must end in ".py"
                # 2. Must contain only valid characters (alphanumeric and underscore)
                # 3. Must start with an alphabetic character
                mod_basename = fname[:-3]
                if not re.match(r"[a-zA-Z][a-zA-Z0-9_]*$", mod_basename):
                    continue

                # Check if the file contains the word "_get_packed_fingerprint"
                full_fname = os.path.join(root, fname)
                try:
                    contents = open(full_fname, "r", encoding="latin1").read()
                except IOError:
                    continue
                if not regex.search(contents):
                    continue

                # Check to see if the file corresponds to a LCM type module genereated by lcm-gen. Parse the
                # file using pyclbr, and check if it contains a class with the right name and methods
                if python_package:
                    modname = "%s.%s" % (python_package, mod_basename)
                else:
                    modname = mod_basename
                try:
                    klass = pyclbr.readmodule(modname).get(mod_basename)
                    if klass and all(
                        method in klass.methods
                        for method in ["decode", "_get_packed_fingerprint"]
                    ):
                        lcmtypes.append(modname)
                except ImportError:
                    continue

            # Only recurse into subdirectories that correspond to python packages (i.e., they contain a file named
            # "__init__.py")
            subdirs_to_traverse = [
                subdir_name
                for subdir_name in dirs
                if os.path.exists(os.path.join(root, subdir_name, "__init__.py"))
            ]
            del dirs[:]
            dirs.extend(subdirs_to_traverse)
    return lcmtypes


def make_lcmtype_dictionary(
    lcm_packages: List[str] = [],
) -> Dict[Union[bytes, bytearray], Any]:
    """
    Create a dictionary of LCM types keyed by fingerprint.

    Searches the specified python package directories for modules
    corresponding to LCM types, imports all the discovered types into the
    global namespace, and returns a dictionary mapping packed fingerprints
    to LCM type classes.

    The primary use for this dictionary is to automatically identify and
    decode an LCM message.

    Args:
        lcm_packages (List[str]): List of directories to search for LCM types in
        addition to the directories in the python path.

    Returns:
        lcm_objects (Dict[Union[bytes, bytearray], Any]): LCM objects found on
        the specified python package directories keyed by fingerprint.
    """
    lcmtypes = find_lcmtypes(lcm_packages=lcm_packages)

    result = {}

    for lcmtype_name in lcmtypes:
        try:
            __import__(lcmtype_name)
            mod = sys.modules[lcmtype_name]
            type_basename = lcmtype_name.split(".")[-1]
            klass = getattr(mod, type_basename)
            fingerprint = klass._get_packed_fingerprint()
            result[fingerprint] = klass
        except ImportError:
            print("Error importing %s" % lcmtype_name)
    return result


if __name__ == "__main__":
    import binascii

    print("Searching for LCM types...")
    lcmtypes = make_lcmtype_dictionary()
    num_types = len(lcmtypes)
    print(f"Found {num_types} type{'s' if num_types != 1 else ''}")
    for fingerprint, klass in lcmtypes.items():
        print(binascii.hexlify(fingerprint), klass.__module__)



"""
Test module for the vector math module.

Authors: Sebastián Rodríguez-Martínez
Contact: srodriguez@mbari.org
Last Update: 2024-03-25
"""

import numpy as np

from navlib import (
    derivative,
    difference,
    distance_traveled,
    max,
    mean,
    median,
    min,
    norm,
    normalize,
    print_stats,
    remove_mean,
    remove_median,
    remove_offset,
    resample,
    saturate,
    std,
    transpose,
    unwrap1pi,
    unwrap2pi,
    unwrap180,
    unwrap360,
    wrap1pi,
    wrap2pi,
    wrap180,
    wrap360,
    wrapunwrap,
    wrapunwrap360,
)
from navlib.math import derivative as derivative_relative
from navlib.math import difference as difference_relative
from navlib.math import distance_traveled as distance_traveled_relative
from navlib.math import max as max_relative
from navlib.math import mean as mean_relative
from navlib.math import median as median_relative
from navlib.math import min as min_relative
from navlib.math import norm as norm_relative
from navlib.math import normalize as normalize_relative
from navlib.math import print_stats as print_stats_relative
from navlib.math import remove_mean as remove_mean_relative
from navlib.math import remove_median as remove_median_relative
from navlib.math import remove_offset as remove_offset_relative
from navlib.math import resample as resample_relative
from navlib.math import saturate as saturate_relative
from navlib.math import std as std_relative
from navlib.math import transpose as transpose_relative
from navlib.math import unwrap1pi as unwrap1pi_relative
from navlib.math import unwrap2pi as unwrap2pi_relative
from navlib.math import unwrap180 as unwrap180_relative
from navlib.math import unwrap360 as unwrap360_relative
from navlib.math import wrap1pi as wrap1pi_relative
from navlib.math import wrap2pi as wrap2pi_relative
from navlib.math import wrap180 as wrap180_relative
from navlib.math import wrap360 as wrap360_relative
from navlib.math import wrapunwrap as wrapunwrap_relative
from navlib.math import wrapunwrap360 as wrapunwrap360_relative


def test_imports():
    # Checks that the absolute and relative imports work correctly
    assert derivative_relative.__code__.co_code == derivative.__code__.co_code
    assert difference_relative.__code__.co_code == difference.__code__.co_code
    assert (
        distance_traveled_relative.__code__.co_code
        == distance_traveled.__code__.co_code
    )
    assert max_relative.__code__.co_code == max.__code__.co_code
    assert mean_relative.__code__.co_code == mean.__code__.co_code
    assert median_relative.__code__.co_code == median.__code__.co_code
    assert min_relative.__code__.co_code == min.__code__.co_code
    assert norm_relative.__code__.co_code == norm.__code__.co_code
    assert normalize_relative.__code__.co_code == normalize.__code__.co_code
    assert print_stats_relative.__code__.co_code == print_stats.__code__.co_code
    assert remove_mean_relative.__code__.co_code == remove_mean.__code__.co_code
    assert remove_median_relative.__code__.co_code == remove_median.__code__.co_code
    assert remove_offset_relative.__code__.co_code == remove_offset.__code__.co_code
    assert resample_relative.__code__.co_code == resample.__code__.co_code
    assert saturate_relative.__code__.co_code == saturate.__code__.co_code
    assert std_relative.__code__.co_code == std.__code__.co_code
    assert transpose_relative.__code__.co_code == transpose.__code__.co_code
    assert unwrap1pi_relative.__code__.co_code == unwrap1pi.__code__.co_code
    assert unwrap2pi_relative.__code__.co_code == unwrap2pi.__code__.co_code
    assert unwrap180_relative.__code__.co_code == unwrap180.__code__.co_code
    assert unwrap360_relative.__code__.co_code == unwrap360.__code__.co_code
    assert wrap1pi_relative.__code__.co_code == wrap1pi.__code__.co_code
    assert wrap2pi_relative.__code__.co_code == wrap2pi.__code__.co_code
    assert wrap180_relative.__code__.co_code == wrap180.__code__.co_code
    assert wrap360_relative.__code__.co_code == wrap360.__code__.co_code
    assert wrapunwrap_relative.__code__.co_code == wrapunwrap.__code__.co_code
    assert wrapunwrap360_relative.__code__.co_code == wrapunwrap360.__code__.co_code


def test_norm():
    """
    Test the norm function.
    """
    # Test case 1: 1D vector
    vec = np.array([1, 2, 3])
    expected_norm = 3.7416573867739413
    assert norm(vec) - expected_norm < 1e-6

    # Test case 2: 2D vector with keepdims
    vec = np.array([[1, 1, 1], [2, 2, 2], [3, 3, 3]])
    expected_norm = np.array([[1.73205081], [3.46410162], [5.19615242]])
    assert np.allclose(norm(vec, keepdims=True), expected_norm)

    # Test case 3: 2D vector without keepdims
    vec = np.array([[1, 1, 1], [2, 2, 2], [3, 3, 3]])
    expected_norm = np.array([1.73205081, 3.46410162, 5.19615242])
    assert np.allclose(norm(vec, keepdims=False), expected_norm)

    # Test case 4: List input
    vec = [1, 2, 3]
    expected_norm = 3.7416573867739413
    assert norm(vec) - expected_norm < 1e-6

    # Test case 5: Wrong input type
    vec = "1, 2, 3"
    try:
        norm(vec)
    except TypeError as e:
        assert str(e) == "Input must be a numpy array or a list."
    else:
        assert False


def test_normalize():
    """
    Test the normalize function.
    """
    # Test case 1: 1D vector
    vec = np.array([1, 2, 3])
    expected_normalized = np.array([0.26726124, 0.53452248, 0.80178373])
    assert np.allclose(normalize(vec), expected_normalized)

    # Test case 2: 2D vector
    vec = np.array([[1, 1, 1], [2, 2, 2], [3, 3, 3]])
    expected_normalized = np.array(
        [
            [0.57735027, 0.57735027, 0.57735027],
            [0.57735027, 0.57735027, 0.57735027],
            [0.57735027, 0.57735027, 0.57735027],
        ]
    )
    assert np.allclose(normalize(vec), expected_normalized)

    # Test case 3: List input
    vec = [1, 2, 3]
    expected_normalized = np.array([0.26726124, 0.53452248, 0.80178373])
    assert np.allclose(normalize(vec), expected_normalized)

    # Test case 4: Wrong input type
    vec = "1, 2, 3"
    try:
        normalize(vec)
    except TypeError as e:
        assert str(e) == "Input must be a numpy array or a list."
    else:
        assert False


def test_remove_offset():
    """
    Test the remove_offset function.
    """
    # Test case 1: 1D vector
    vec = np.array([1, 2, 3])
    offset = 1
    expected_vec = np.array([0, 1, 2])
    assert np.allclose(remove_offset(vec, offset), expected_vec)

    # Test case 2: 2D vector
    vec = np.array([[1, 1, 1], [2, 2, 2], [3, 3, 3]])
    offset = 1
    expected_vec = np.array([[0, 0, 0], [1, 1, 1], [2, 2, 2]])
    assert np.allclose(remove_offset(vec, offset), expected_vec)

    # Test case 3: List input
    vec = [1, 2, 3]
    offset = 1
    expected_vec = np.array([0, 1, 2])
    assert np.allclose(remove_offset(vec, offset), expected_vec)

    # Test case 4: Wrong input type
    vec = "1, 2, 3"
    offset = 1
    try:
        remove_offset(vec, offset)
    except TypeError as e:
        assert str(e) == "Input must be a numpy array or a list."
    else:
        assert False


def test_remove_mean():
    """
    Test the remove_mean function.
    """
    # Test case 1: 1D vector
    vec = np.array([1, 2, 3])
    expected_vec = np.array([-1, 0, 1])
    assert np.allclose(remove_mean(vec), expected_vec)

    # Test case 2: 2D vector
    vec = np.array([[1, 1, 1], [2, 2, 2], [3, 3, 3]])
    expected_vec = np.array([[-1, -1, -1], [0, 0, 0], [1, 1, 1]])
    assert np.allclose(remove_mean(vec), expected_vec)

    # Test case 3: List input
    vec = [1, 2, 3]
    expected_vec = np.array([-1, 0, 1])
    assert np.allclose(remove_mean(vec), expected_vec)

    # Test case 4: Wrong input type
    vec = "1, 2, 3"
    try:
        remove_mean(vec)
    except TypeError as e:
        assert str(e) == "Input must be a numpy array or a list."
    else:
        assert False


def test_remove_median():
    """
    Test the remove_median function.
    """
    # Test case 1: 1D vector
    vec = np.array([1, 2, 3])
    expected_vec = np.array([-1, 0, 1])
    assert np.allclose(remove_median(vec), expected_vec)

    # Test case 2: 2D vector
    vec = np.array([[1, 1, 1], [2, 2, 2], [3, 3, 3]])
    expected_vec = np.array([[-1, -1, -1], [0, 0, 0], [1, 1, 1]])
    assert np.allclose(remove_median(vec), expected_vec)

    # Test case 3: List input
    vec = [1, 2, 3]
    expected_vec = np.array([-1, 0, 1])
    assert np.allclose(remove_median(vec), expected_vec)

    # Test case 4: Wrong input type
    vec = "1, 2, 3"
    try:
        remove_median(vec)
    except TypeError as e:
        assert str(e) == "Input must be a numpy array or a list."
    else:
        assert False


def test_mean():
    """
    Test the mean function.
    """
    # Test case 1: 1D vector
    vec = np.array([1, 2, 3])
    expected_mean = 2
    assert mean(vec) == expected_mean

    # Test case 2: 2D vector withou keepdims
    vec = np.array([[1, 1, 1], [2, 2, 2], [3, 3, 3]])
    expected_mean = np.array([2, 2, 2])
    assert np.allclose(mean(vec), expected_mean)

    # Test case 3: 2D vector with keepdims
    vec = np.array([[1, 1, 1], [2, 2, 2], [3, 3, 3]])
    expected_mean = np.array([[2, 2, 2]])
    assert np.allclose(mean(vec, keepdims=True), expected_mean)

    # Test case 4: List input
    vec = [1, 2, 3]
    expected_mean = 2
    assert mean(vec) == expected_mean

    # Test case 5: Wrong input type
    vec = "1, 2, 3"
    try:
        mean(vec)
    except TypeError as e:
        assert str(e) == "Input must be a numpy array or a list."
    else:
        assert False


def test_median():
    """
    Test the median function.
    """
    # Test case 1: 1D vector
    vec = np.array([1, 2, 3])
    expected_median = 2
    assert median(vec) == expected_median

    # Test case 2: 2D vector withou keepdims
    vec = np.array([[1, 1, 1], [2, 2, 2], [3, 3, 3]])
    expected_median = np.array([2, 2, 2])
    assert np.allclose(median(vec), expected_median)

    # Test case 3: 2D vector with keepdims
    vec = np.array([[1, 1, 1], [2, 2, 2], [3, 3, 3]])
    expected_median = np.array([[2, 2, 2]])
    assert np.allclose(median(vec, keepdims=True), expected_median)

    # Test case 4: List input
    vec = [1, 2, 3]
    expected_median = 2
    assert median(vec) == expected_median

    # Test case 5: Wrong input type
    vec = "1, 2, 3"
    try:
        median(vec)
    except TypeError as e:
        assert str(e) == "Input must be a numpy array or a list."
    else:
        assert False


def test_std():
    """
    Test the std function.
    """
    # Test case 1: 1D vector
    vec = np.array([1, 2, 3])
    expected_std = 0.816496580927726
    assert std(vec) - expected_std < 1e-6

    # Test case 2: 2D vector without keepdims
    vec = np.array([[1, 1, 1], [2, 2, 2], [3, 3, 3]])
    expected_std = np.array([0.816496580927726, 0.816496580927726, 0.816496580927726])
    assert np.allclose(std(vec), expected_std)

    # Test case 3: 2D vector with keepdims
    vec = np.array([[1, 1, 1], [2, 2, 2], [3, 3, 3]])
    expected_std = np.array([[0.816496580927726, 0.816496580927726, 0.816496580927726]])
    assert np.allclose(std(vec, keepdims=True), expected_std)

    # Test case 4: List input
    vec = [1, 2, 3]
    expected_std = 0.816496580927726
    assert std(vec) - expected_std < 1e-6

    # Test case 5: Wrong input type
    vec = "1, 2, 3"
    try:
        std(vec)
    except TypeError as e:
        assert str(e) == "Input must be a numpy array or a list."
    else:
        assert False


def test_min():
    """
    Test the min function.
    """
    # Test case 1: 1D vector
    vec = np.array([1, 2, 3])
    expected_min = 1
    assert min(vec) == expected_min

    # Test case 2: 2D vector withou keepdims
    vec = np.array([[1, 1, 1], [2, 2, 2], [3, 3, 3]])
    expected_min = np.array([1, 1, 1])
    assert np.allclose(min(vec), expected_min)

    # Test case 3: 2D vector with keepdims
    vec = np.array([[1, 1, 1], [2, 2, 2], [3, 3, 3]])
    expected_min = np.array([[1, 1, 1]])
    assert np.allclose(min(vec, keepdims=True), expected_min)

    # Test case 4: List input
    vec = [1, 2, 3]
    expected_min = 1
    assert min(vec) == expected_min

    # Test case 5: Wrong input type
    vec = "1, 2, 3"
    try:
        min(vec)
    except TypeError as e:
        assert str(e) == "Input must be a numpy array or a list."
    else:
        assert False


def test_max():
    """
    Test the max function.
    """
    # Test case 1: 1D vector
    vec = np.array([1, 2, 3])
    expected_max = 3
    assert max(vec) == expected_max

    # Test case 2: 2D vector withou keepdims
    vec = np.array([[1, 1, 1], [2, 2, 2], [3, 3, 3]])
    expected_max = np.array([3, 3, 3])
    assert np.allclose(max(vec), expected_max)

    # Test case 3: 2D vector with keepdims
    vec = np.array([[1, 1, 1], [2, 2, 2], [3, 3, 3]])
    expected_max = np.array([[3, 3, 3]])
    assert np.allclose(max(vec, keepdims=True), expected_max)

    # Test case 4: List input
    vec = [1, 2, 3]
    expected_max = 3
    assert max(vec) == expected_max

    # Test case 5: Wrong input type
    vec = "1, 2, 3"
    try:
        max(vec)
    except TypeError as e:
        assert str(e) == "Input must be a numpy array or a list."
    else:
        assert False


def test_print_stats():
    """
    Test the print_stats function.
    """
    # Test case 1: 1D vector
    vec = np.array([1, 2, 3])
    print_stats(vec)

    # Test case 2: 2D vector
    vec = np.array([[1, 1, 1], [2, 2, 2], [3, 3, 3]])
    print_stats(vec)

    # Test case 3: List input
    vec = [1, 2, 3]
    print_stats(vec)

    # Test case 4: Wrong input type
    vec = "1, 2, 3"
    try:
        print_stats(vec)
    except TypeError as e:
        assert str(e) == "Input must be a numpy array or a list."
    else:
        assert False


def test_difference():
    """
    Test the difference function.
    """
    # Test case 1: 1D vector
    vec = np.array([1, 2, 3])
    expected_diff = np.array([1, 1])
    assert np.allclose(difference(vec), expected_diff)

    # Test case 2: 2D vector
    vec = np.array([[1, 1, 1], [2, 2, 2], [3, 3, 3]])
    expected_diff = np.array([[1, 1, 1], [1, 1, 1]])
    assert np.allclose(difference(vec), expected_diff)

    # Test case 3: List input
    vec = [1, 2, 3]
    expected_diff = np.array([1, 1])
    assert np.allclose(difference(vec), expected_diff)

    # Test case 4: Wrong input type
    vec = "1, 2, 3"
    try:
        difference(vec)
    except TypeError as e:
        assert str(e) == "Input must be a numpy array or a list."
    else:
        assert False


def test_derivative():
    """
    Test the derivative function.
    """
    # Test case 1: 1D vector
    vec = np.array([1, 2, 3])
    time = np.array([0.1, 0.2, 0.3])
    expected_deriv = np.array([10, 10])
    assert np.allclose(derivative(vec, time), expected_deriv)

    # Test case 2: 2D vector
    vec = np.array([[1, 1, 1], [2, 2, 2], [3, 3, 3]])
    time = np.array([0.1, 0.2, 0.3])
    expected_deriv = np.array([[10, 10, 10], [10, 10, 10]])
    assert np.allclose(derivative(vec, time), expected_deriv)

    # Test case 3: List input
    vec = [1, 2, 3]
    time = [0.1, 0.2, 0.3]
    expected_deriv = np.array([10, 10])
    assert np.allclose(derivative(vec, time), expected_deriv)

    # Test case 4: Wrong input type
    vec = "1, 2, 3"
    time = [0.1, 0.2, 0.3]
    try:
        derivative(vec, time)
    except TypeError as e:
        assert str(e) == "Input must be a numpy array or a list."
    else:
        assert False


def test_resample():
    """
    Test the resample function.
    """
    # Test case 1: 1D vector
    vec = np.array([1, 2, 3, 4])
    time = np.array([0.1, 0.2, 0.3, 0.4])
    new_time = np.array([0.15, 0.25, 0.35])
    expected_resampled = np.array([1.5, 2.5, 3.5])
    assert np.allclose(resample(new_time, time, vec), expected_resampled)

    # Test case 2: 2D vector
    vec = np.array([[1, 1, 1], [2, 2, 2], [3, 3, 3], [4, 4, 4]])
    time = np.array([0.1, 0.2, 0.3, 0.4])
    new_time = np.array([0.15, 0.25, 0.35])
    expected_resampled = np.array([[1.5, 1.5, 1.5], [2.5, 2.5, 2.5], [3.5, 3.5, 3.5]])
    assert np.allclose(resample(new_time, time, vec), expected_resampled)

    # Test case 3: List input
    vec = [1, 2, 3, 4]
    time = [0.1, 0.2, 0.3, 0.4]
    new_time = [0.15, 0.25, 0.35]
    expected_resampled = np.array([1.5, 2.5, 3.5])
    assert np.allclose(resample(new_time, time, vec), expected_resampled)

    # Test case 4: Wrong input type
    vec_str = "1, 2, 3, 4"
    vec = [1, 2, 3, 4]
    time_str = "0.1, 0.2, 0.3, 0.4"
    time_array = np.array([[0.1, 0.2, 0.3, 0.4], [0.1, 0.2, 0.3, 0.4]])
    time_longer = [0.1, 0.2, 0.3, 0.4, 0.5]
    time_non_monotonic = [0.1, 0.2, 0.18, 0.4]
    time = [0.1, 0.2, 0.3, 0.4]
    new_time_str = "0.15, 0.25, 0.35"
    new_time = [0.15, 0.25, 0.35]
    try:
        resample(new_time, time, vec_str)
    except TypeError as e:
        assert str(e) == "Input fp must be a numpy array or a list."
    else:
        assert False
    try:
        resample(new_time_str, time, vec)
    except TypeError as e:
        assert str(e) == "Input x must be a numpy array or a list."
    else:
        assert False
    try:
        resample(new_time, time_str, vec)
    except TypeError as e:
        assert str(e) == "Input xp must be a numpy array or a list."
    else:
        assert False
    try:
        resample(new_time, time_array, vec)
    except ValueError as e:
        assert str(e) == "Input xp must be a 1-D sequence of n floats."
    else:
        assert False
    try:
        resample(new_time, time_longer, vec)
    except ValueError as e:
        assert (
            str(e)
            == "The number of samples in the old time vector and the signal must match."
        )
    else:
        assert False
    try:
        resample(new_time, time_non_monotonic, vec)
    except ValueError as e:
        assert str(e) == "Time series must be monotonically increasing."
    else:
        assert False


def test_distance_traveled():
    """
    Test the distance_traveled function
    """
    # Test case 1: Linear and Full 2D distance
    mat = np.array([[0, 0], [1, 1], [2, 2]])
    expected = np.linalg.norm(mat, axis=1)
    result = distance_traveled(mat, "2d", True, True)
    np.testing.assert_array_equal(result, expected)

    # Test case 2: Linear and non-full 2D distance
    mat = np.array([[0, 0], [1, 1], [2, 2]])
    expected = 2 * np.sqrt(2)
    result = distance_traveled(mat, "2d", True, False)
    assert result == expected

    # Test case 3: non-linear and full 2D distance
    mat = np.array([[0, 0], [1, 1], [2, 2]])
    expected = np.r_[0.0, np.cumsum(np.linalg.norm(np.diff(mat, axis=0), axis=1))]
    result = distance_traveled(mat, "2d", False, True)
    np.testing.assert_array_equal(result, expected)

    # Test case 4: non-linear and non-full 2D distance
    mat = np.array([[0, 0], [1, 1], [2, 2]])
    expected = 2 * np.sqrt(2)
    result = distance_traveled(mat, "2d", False, False)
    assert result == expected

    # Test case 5: Linear and Full 3D distance
    mat = np.array([[0, 0, 0], [1, 1, 1], [2, 2, 2]])
    expected = np.linalg.norm(mat, axis=1)
    result = distance_traveled(mat, "3d", True, True)
    np.testing.assert_array_equal(result, expected)

    # Test case 6: Linear and non-full 3D distance
    mat = np.array([[0, 0, 0], [1, 1, 1], [2, 2, 2]])
    expected = 2 * np.sqrt(3)
    result = distance_traveled(mat, "3d", True, False)
    assert result == expected

    # Test case 7: non-linear and full 3D distance
    mat = np.array([[0, 0, 0], [1, 1, 1], [2, 2, 2]])
    expected = np.r_[0.0, np.cumsum(np.linalg.norm(np.diff(mat, axis=0), axis=1))]
    result = distance_traveled(mat, "3d", False, True)
    np.testing.assert_array_equal(result, expected)

    # Test case 8: non-linear and non-full 3D distance
    mat = np.array([[0, 0, 0], [1, 1, 1], [2, 2, 2]])
    expected = 2 * np.sqrt(3)
    result = distance_traveled(mat, "3d", False, False)
    assert result == expected

    # Test case 9: Incorrect dimensions
    mat = np.array([[0, 0], [1, 1], [2, 2]])
    try:
        distance_traveled(mat, "4d")
    except ValueError as e:
        assert str(e) == "Dimensions must be either 2d or 3d."
    else:
        assert False

    # Test case 10: Insufficient columns
    mat = np.array([[0], [1], [2]])
    try:
        distance_traveled(mat, "2d")
    except ValueError as e:
        assert str(e) == "The input matrix must have at least two columns."
    else:
        assert False


def test_saturate():
    """
    Test the saturate function
    """
    # Test case 1: 1D vector
    vec = np.array([1, 2, 3])
    lower = 1.5
    upper = 2.5
    expected_saturated = np.array([1.5, 2, 2.5])
    assert np.allclose(saturate(vec, lower, upper), expected_saturated)

    # Test case 2: 2D vector
    vec = np.array([[1, 1, 1], [2, 2, 2], [3, 3, 3]])
    lower = 1.5
    upper = 2.5
    expected_saturated = np.array([[1.5, 1.5, 1.5], [2, 2, 2], [2.5, 2.5, 2.5]])
    assert np.allclose(saturate(vec, lower, upper), expected_saturated)

    # Test case 3: List input
    vec = [1, 2, 3]
    lower = 1.5
    upper = 2.5
    expected_saturated = np.array([1.5, 2, 2.5])
    assert np.allclose(saturate(vec, lower, upper), expected_saturated)

    # Test case 4: Wrong input type
    vec = "1, 2, 3"
    lower = 1.5
    upper = 2.5
    try:
        saturate(vec, lower, upper)
    except TypeError as e:
        assert str(e) == "Input must be a numpy array or a list."
    else:
        assert False


def test_wrap360():
    """
    Test the wrap360 function.
    """
    # Test case 1: Unidimensional array
    angles = np.array(
        [
            -400,
            -180,
            -120,
            -90,
            -60,
            -45,
            -30,
            -10,
            0,
            10,
            30,
            45,
            60,
            90,
            120,
            180,
            400,
        ]
    )
    expected_angles = np.array(
        [320, 180, 240, 270, 300, 315, 330, 350, 0, 10, 30, 45, 60, 90, 120, 180, 40]
    )
    assert np.allclose(wrap360(angles), expected_angles)
    assert np.allclose(wrap360(angles.tolist()), expected_angles)

    # Test case 2: Multidimensional array
    angles = np.array(
        [
            [
                -400,
                -180,
                -120,
                -90,
                -60,
                -45,
                -30,
                -10,
                0,
                10,
                30,
                45,
                60,
                90,
                120,
                180,
                400,
            ],
            [
                -400,
                -180,
                -120,
                -90,
                -60,
                -45,
                -30,
                -10,
                0,
                10,
                30,
                45,
                60,
                90,
                120,
                180,
                400,
            ],
        ]
    ).T
    expected_angles = np.array(
        [
            [
                320,
                180,
                240,
                270,
                300,
                315,
                330,
                350,
                0,
                10,
                30,
                45,
                60,
                90,
                120,
                180,
                40,
            ],
            [
                320,
                180,
                240,
                270,
                300,
                315,
                330,
                350,
                0,
                10,
                30,
                45,
                60,
                90,
                120,
                180,
                40,
            ],
        ]
    ).T
    assert np.allclose(wrap360(angles), expected_angles)
    assert np.allclose(wrap360(angles.tolist()), expected_angles)
    assert np.allclose(wrap360(angles.T), expected_angles.T)
    assert np.allclose(wrap360(angles.T.tolist()), expected_angles.T)


def test_wrap180():
    """
    Test the wrap180 function.
    """
    # Test case 1: Unidimensional array
    angles = np.array([-360, -270, -180, -90, 0, 90, 180, 270, 360])
    expected_angles = np.array([0, 90, -180, -90, 0, 90, -180, -90, 0])
    assert np.allclose(wrap180(angles), expected_angles)
    assert np.allclose(wrap180(angles.tolist()), expected_angles)

    # Test case 2: Multidimensional array
    angles = np.array(
        [
            [-360, -270, -180, -90, 0, 90, 180, 270, 360],
            [-360, -270, -180, -90, 0, 90, 180, 270, 360],
        ]
    ).T
    expected_angles = np.array(
        [
            [0, 90, -180, -90, 0, 90, -180, -90, 0],
            [0, 90, -180, -90, 0, 90, -180, -90, 0],
        ]
    ).T
    assert np.allclose(wrap180(angles), expected_angles)
    assert np.allclose(wrap180(angles.tolist()), expected_angles)
    assert np.allclose(wrap180(angles.T), expected_angles.T)
    assert np.allclose(wrap180(angles.T.tolist()), expected_angles.T)


def test_wrap2pi():
    """
    Test the wrap2pi function.
    """
    # Test case 1: Unidimensional array
    angles = np.deg2rad(
        np.array(
            [
                -400,
                -180,
                -120,
                -90,
                -60,
                -45,
                -30,
                -10,
                0,
                10,
                30,
                45,
                90,
                120,
                180,
                400,
            ]
        )
    )
    expected_angles = np.deg2rad(
        np.array(
            [320, 180, 240, 270, 300, 315, 330, 350, 0, 10, 30, 45, 90, 120, 180, 40]
        )
    )
    assert np.allclose(wrap2pi(angles), expected_angles)
    assert np.allclose(wrap2pi(angles.tolist()), expected_angles)

    # Test case 2: Multidimensional array
    angles = np.deg2rad(
        np.array(
            [
                [
                    -400,
                    -180,
                    -120,
                    -90,
                    -60,
                    -45,
                    -30,
                    -10,
                    0,
                    10,
                    30,
                    45,
                    90,
                    120,
                    180,
                ],
                [
                    -400,
                    -180,
                    -120,
                    -90,
                    -60,
                    -45,
                    -30,
                    -10,
                    0,
                    10,
                    30,
                    45,
                    90,
                    120,
                    180,
                ],
            ]
        ).T
    )
    expected_angles = np.deg2rad(
        np.array(
            [
                [320, 180, 240, 270, 300, 315, 330, 350, 0, 10, 30, 45, 90, 120, 180],
                [320, 180, 240, 270, 300, 315, 330, 350, 0, 10, 30, 45, 90, 120, 180],
            ]
        ).T
    )
    assert np.allclose(wrap2pi(angles), expected_angles)
    assert np.allclose(wrap2pi(angles.tolist()), expected_angles)
    assert np.allclose(wrap2pi(angles.T), expected_angles.T)
    assert np.allclose(wrap2pi(angles.T.tolist()), expected_angles.T)


def test_wrap1pi():
    """
    Test the wrap1pi function.
    """
    # Test case 1: Unidimensional array
    angles = np.deg2rad(np.array([-360, -270, -180, -90, 0, 90, 180, 270, 360]))
    expected_angles = np.deg2rad(np.array([0, 90, -180, -90, 0, 90, -180, -90, 0]))
    assert np.allclose(wrap1pi(angles), expected_angles)
    assert np.allclose(wrap1pi(angles.tolist()), expected_angles)

    # Test case 2: Multidimensional array
    angles = np.deg2rad(
        np.array(
            [
                [-360, -270, -180, -90, 0, 90, 180, 270, 360],
                [-360, -270, -180, -90, 0, 90, 180, 270, 360],
            ]
        ).T
    )
    expected_angles = np.deg2rad(
        np.array(
            [
                [0, 90, -180, -90, 0, 90, -180, -90, 0],
                [0, 90, -180, -90, 0, 90, -180, -90, 0],
            ]
        ).T
    )
    assert np.allclose(wrap1pi(angles), expected_angles)
    assert np.allclose(wrap1pi(angles.tolist()), expected_angles)
    assert np.allclose(wrap1pi(angles.T), expected_angles.T)
    assert np.allclose(wrap1pi(angles.T.tolist()), expected_angles.T)


def test_unwrap2pi():
    """
    Test the unwrap2pi function.
    """
    # Test case 1: No discontinuities
    angles = np.deg2rad(
        np.array([0, 90, 180, 270, 360, 90, 180, 270, 360, 90, 180, 270, 360])
    )
    expected_angles = np.deg2rad(
        np.array([0, 90, 180, 270, 360, 450, 540, 630, 720, 810, 900, 990, 1080])
    )
    assert np.allclose(unwrap2pi(angles), expected_angles)
    assert np.allclose(unwrap2pi(angles.tolist()), expected_angles)

    # Test case 2: Discontinuities
    angles = np.deg2rad(np.array([0, 90, 180, 270, 360, 90, 320, 350, 380]))
    expected_angles = np.deg2rad(np.array([0, 90, 180, 270, 360, 450, 320, 350, 380]))
    assert np.allclose(unwrap2pi(angles), expected_angles)
    assert np.allclose(unwrap2pi(angles.tolist()), expected_angles)

    # Test case 3: Multidimensional array
    angles = np.deg2rad(
        np.array(
            [
                [0, 90, 180, 270, 360, 90, 180, 270, 360, 90, 180, 270, 360],
                [0, 90, 180, 270, 360, 90, 180, 270, 360, 90, 180, 270, 360],
            ]
        ).T
    )
    expected_angles = np.deg2rad(
        np.array(
            [
                [0, 90, 180, 270, 360, 450, 540, 630, 720, 810, 900, 990, 1080],
                [0, 90, 180, 270, 360, 450, 540, 630, 720, 810, 900, 990, 1080],
            ]
        ).T
    )
    assert np.allclose(unwrap2pi(angles), expected_angles)
    assert np.allclose(unwrap2pi(angles.tolist()), expected_angles)


def test_unwrap1pi():
    """
    Test the unwrap1pi function.
    """
    # Test case 1: No discontinuities
    angles = np.deg2rad(
        np.array([0, 45, 90, 135, 180, 45, 90, 135, 180, 45, 90, 135, 180])
    )
    expected_angles = np.deg2rad(
        np.array([0, 45, 90, 135, 180, 225, 270, 315, 360, 405, 450, 495, 540])
    )
    assert np.allclose(unwrap1pi(angles), expected_angles)
    assert np.allclose(unwrap1pi(angles.tolist()), expected_angles)

    # Test case 2: Discontinuities
    angles = np.deg2rad(np.array([0, 45, 90, 135, 180, 45, 145, 170, 180]))
    expected_angles = np.deg2rad(np.array([0, 45, 90, 135, 180, 225, 145, 170, 180]))
    assert np.allclose(unwrap1pi(angles), expected_angles)
    assert np.allclose(unwrap1pi(angles.tolist()), expected_angles)

    # Test case 3: Multidimensional array
    angles = np.deg2rad(
        np.array(
            [
                [0, 45, 90, 135, 180, 45, 90, 135, 180, 45, 90, 135, 180],
                [0, 45, 90, 135, 180, 45, 90, 135, 180, 45, 90, 135, 180],
            ]
        ).T
    )
    expected_angles = np.deg2rad(
        np.array(
            [
                [0, 45, 90, 135, 180, 225, 270, 315, 360, 405, 450, 495, 540],
                [0, 45, 90, 135, 180, 225, 270, 315, 360, 405, 450, 495, 540],
            ]
        ).T
    )
    assert np.allclose(unwrap1pi(angles), expected_angles)
    assert np.allclose(unwrap1pi(angles.tolist()), expected_angles)


def test_unwrap360():
    """
    Test the unwrap360 function.
    """
    # Test case 1: No discontinuities
    angles = np.array([0, 90, 180, 270, 360, 90, 180, 270, 360, 90, 180, 270, 360])
    expected_angles = np.array(
        [0, 90, 180, 270, 360, 450, 540, 630, 720, 810, 900, 990, 1080]
    )
    assert np.allclose(unwrap360(angles), expected_angles)
    assert np.allclose(unwrap360(angles.tolist()), expected_angles)

    # Test case 2: Discontinuities
    angles = np.array([0, 90, 180, 270, 360, 90, 320, 350, 380])
    expected_angles = np.array([0, 90, 180, 270, 360, 450, 320, 350, 380])
    assert np.allclose(unwrap360(angles), expected_angles)
    assert np.allclose(unwrap360(angles.tolist()), expected_angles)

    # Test case 3: Multidimensional array
    angles = np.array(
        [
            [0, 90, 180, 270, 360, 90, 180, 270, 360, 90, 180, 270, 360],
            [0, 90, 180, 270, 360, 90, 180, 270, 360, 90, 180, 270, 360],
        ]
    ).T
    expected_angles = np.array(
        [
            [0, 90, 180, 270, 360, 450, 540, 630, 720, 810, 900, 990, 1080],
            [0, 90, 180, 270, 360, 450, 540, 630, 720, 810, 900, 990, 1080],
        ]
    ).T
    assert np.allclose(unwrap360(angles), expected_angles)
    assert np.allclose(unwrap360(angles.tolist()), expected_angles)


def test_unwrap180():
    """
    Test the unwrap180 function.
    """
    # Test case 1: No discontinuities
    angles = np.array([0, 45, 90, 135, 180, 45, 90, 135, 180, 45, 90, 135, 180])
    expected_angles = np.array(
        [0, 45, 90, 135, 180, 225, 270, 315, 360, 405, 450, 495, 540]
    )
    assert np.allclose(unwrap180(angles), expected_angles)
    assert np.allclose(unwrap180(angles.tolist()), expected_angles)

    # Test case 2: Discontinuities
    angles = np.array([0, 45, 90, 135, 180, 45, 145, 170, 180])
    expected_angles = np.array([0, 45, 90, 135, 180, 225, 145, 170, 180])
    assert np.allclose(unwrap180(angles), expected_angles)
    assert np.allclose(unwrap180(angles.tolist()), expected_angles)

    # Test case 3: Multidimensional array
    angles = np.array(
        [
            [0, 45, 90, 135, 180, 45, 90, 135, 180, 45, 90, 135, 180],
            [0, 45, 90, 135, 180, 45, 90, 135, 180, 45, 90, 135, 180],
        ]
    ).T
    expected_angles = np.array(
        [
            [0, 45, 90, 135, 180, 225, 270, 315, 360, 405, 450, 495, 540],
            [0, 45, 90, 135, 180, 225, 270, 315, 360, 405, 450, 495, 540],
        ]
    ).T
    assert np.allclose(unwrap180(angles), expected_angles)
    assert np.allclose(unwrap180(angles.tolist()), expected_angles)


def test_wrapunwrap():
    """
    Test the wrapunwrap function
    """
    # Test case
    angles = np.deg2rad(
        np.array([0, 45, 90, 135, 180, 225, 270, 315, 360, 405, 450, 495, 540])
    )
    expected_angles = np.deg2rad(
        np.array([0, 45, 90, 135, 180, 225, 270, 315, 360, 405, 450, 495, 540])
    )
    assert np.allclose(wrapunwrap(angles), expected_angles)
    assert np.allclose(wrapunwrap(angles.tolist()), expected_angles)

    # Test case 2: Discontinuities
    angles = np.deg2rad(np.array([0, 45, 90, 135, 180, 225, 145, 170, 180]))
    expected_angles = np.deg2rad(np.array([0, 45, 90, 135, 180, 225, 145, 170, 180]))
    assert np.allclose(wrapunwrap(angles), expected_angles)
    assert np.allclose(wrapunwrap(angles.tolist()), expected_angles)

    # Test case 3: Multidimensional array
    angles = np.deg2rad(
        np.array(
            [
                [0, 45, 90, 135, 180, 225, 270, 315, 360, 405, 450, 495, 540],
                [0, 45, 90, 135, 180, 225, 270, 315, 360, 405, 450, 495, 540],
            ]
        ).T
    )
    expected_angles = np.deg2rad(
        np.array(
            [
                [0, 45, 90, 135, 180, 225, 270, 315, 360, 405, 450, 495, 540],
                [0, 45, 90, 135, 180, 225, 270, 315, 360, 405, 450, 495, 540],
            ]
        ).T
    )
    assert np.allclose(wrapunwrap(angles), expected_angles)
    assert np.allclose(wrapunwrap(angles.tolist()), expected_angles)


def test_wrapunwrap360():
    """
    Test the wrapunwrap360 function
    """
    # Test case
    angles = np.array([0, 45, 90, 135, 180, 225, 270, 315, 360, 405, 450, 495, 540])
    expected_angles = np.array(
        [0, 45, 90, 135, 180, 225, 270, 315, 360, 405, 450, 495, 540]
    )
    assert np.allclose(wrapunwrap360(angles), expected_angles)
    assert np.allclose(wrapunwrap360(angles.tolist()), expected_angles)

    # Test case 2: Discontinuities
    angles = np.array([0, 45, 90, 135, 180, 225, 145, 170, 180])
    expected_angles = np.array([0, 45, 90, 135, 180, 225, 145, 170, 180])
    assert np.allclose(wrapunwrap360(angles), expected_angles)
    assert np.allclose(wrapunwrap360(angles.tolist()), expected_angles)

    # Test case 3: Multidimensional array
    angles = np.array(
        [
            [0, 45, 90, 135, 180, 225, 270, 315, 360, 405, 450, 495, 540],
            [0, 45, 90, 135, 180, 225, 270, 315, 360, 405, 450, 495, 540],
        ]
    ).T
    expected_angles = np.array(
        [
            [0, 45, 90, 135, 180, 225, 270, 315, 360, 405, 450, 495, 540],
            [0, 45, 90, 135, 180, 225, 270, 315, 360, 405, 450, 495, 540],
        ]
    ).T
    assert np.allclose(wrapunwrap360(angles), expected_angles)
    assert np.allclose(wrapunwrap360(angles.tolist()), expected_angles)


def test_transpose():
    """
    Test the transpose function.
    """
    # Test case 1: 2D matrix
    mat = np.array([[1, 2], [3, 4]])
    expected_transpose = np.array([[1, 3], [2, 4]])
    assert np.array_equal(transpose(mat), expected_transpose)

    # Test case 2: 3D matrix
    mat = np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])
    expected_transpose = np.array([[[1, 3], [2, 4]], [[5, 7], [6, 8]]])
    assert np.array_equal(transpose(mat), expected_transpose)

    # Test case 3: 1D vector
    vec = np.array([1, 2, 3])
    expected_transpose = np.array([1, 2, 3])  # Transpose of a 1D vector is itself
    assert np.array_equal(transpose(vec), expected_transpose)

    # Test case 4: List input
    mat = [[1, 2], [3, 4]]
    expected_transpose = np.array([[1, 3], [2, 4]])
    assert np.array_equal(transpose(mat), expected_transpose)

    # Test case 5: Wrong input type
    mat = "1, 2, 3"
    try:
        transpose(mat)
    except TypeError as e:
        assert str(e) == "Input must be a numpy array or a list."
    else:
        assert False

    # Test case 6: Dimension greater than 3
    mat = np.random.rand(2, 2, 2, 2)
    try:
        transpose(mat)
    except ValueError as e:
        assert str(e) == "The dimension of the array must be up to 3D"
    else:
        assert False



# import methods from the corresponding modules
from .cal import (
    cal_ahrs_so3,
    cal_dvl_acc_so3,
    cal_dvl_correct_by_sound_velocity,
    cal_dvl_jck_so3,
    cal_mag_ellipsoid_fit,
    cal_mag_ellipsoid_fit_fang,
    cal_mag_magfactor3,
    cal_mag_magyc_bfg,
    cal_mag_magyc_ifg,
    cal_mag_magyc_ls,
    cal_mag_magyc_nls,
    cal_mag_sar_aid,
    cal_mag_sar_kf,
    cal_mag_sar_ls,
    cal_mag_sphere_fit,
    cal_mag_twostep_hi,
    cal_mag_twostep_hsi,
)
from .environment import (
    ctd2oxygen_saturation,
    ctd2salinity,
    ctd2svel,
    ctp2pden,
    ctp2salinity,
    depth2pressure,
)
from .geo import (
    deg2m_lat,
    deg2m_lon,
    dms_to_decimal,
    get_local_gravity,
    latlon_to_zone_letter,
    latlon_to_zone_number,
    ll2utm,
    ll2xy,
    utm2ll,
    xy2ll,
)
from .lcmlog import (
    dict_to_struct,
    filter_channels_from_data,
    lcmlog_to_dict,
    merge_lcm_logs,
    msg_getconstants,
    msg_getfields,
    nav_structtype,
    parse_and_save,
    read_log,
    read_logs,
    struct_to_dict,
)
from .logging import setup_logging
from .math import (
    adjoint,
    axis_ang3,
    axis_ang6,
    derivative,
    difference,
    distance_to_se3,
    distance_to_so3,
    distance_traveled,
    filter_lowpass,
    filter_median,
    filter_savgol,
    hpr2quat,
    hpr2rot,
    left_qmatrix,
    matrix2xyzrph,
    matrix_exp3,
    matrix_exp6,
    matrix_inverse,
    matrix_log3,
    matrix_log6,
    max,
    mean,
    median,
    min,
    norm,
    normalize,
    pose_diff,
    print_stats,
    project_to_se3,
    project_to_so3,
    quat2hpr,
    quat2rot,
    quat2rph,
    quat_conj,
    quat_inv,
    quat_normalize,
    quat_positive_scalar,
    remove_mean,
    remove_median,
    remove_offset,
    resample,
    right_qmatrix,
    rot2hpr,
    rot2quat,
    rot2rph,
    rot_diff,
    rot_inv,
    rph2quat,
    rph2rot,
    saturate,
    se3_to_vec,
    so3_optimization,
    so3_to_vec,
    std,
    transpose,
    unwrap1pi,
    unwrap2pi,
    unwrap180,
    unwrap360,
    vec_to_se3,
    vec_to_so3,
    wrap1pi,
    wrap2pi,
    wrap180,
    wrap360,
    wrapunwrap,
    wrapunwrap360,
    xyzrph2matrix,
)
from .mbsystem import mbs_read_fnv
from .nav import (
    UUV_EKF,
    ahrs_complementary_filter,
    ahrs_correct_magfield,
    ahrs_hua_filter,
    ahrs_madgwick_filter,
    ahrs_mahony_filter,
    ahrs_raw_hdg,
    ahrs_raw_rp,
    ahrs_raw_rph,
    kf_lti_discretize,
    kf_predict,
    kf_update,
    navdvl_dead_reckoning,
    navdvl_kf,
    so3_integrator,
)
from .time import (
    datestr2utime,
    time_interval_indices,
    time_now,
    time_now_utc,
    timestamps_make_uniform,
    utime2datestr,
)

# Get __all__ from the corresponding modules
# time module
__all__ = [
    "time_now",
    "time_now_utc",
    "utime2datestr",
    "datestr2utime",
    "time_interval_indices",
    "timestamps_make_uniform",
]
# nav module
__all__ += [
    "so3_integrator",
    "ahrs_correct_magfield",
    "ahrs_raw_rp",
    "ahrs_raw_hdg",
    "ahrs_raw_rph",
    "ahrs_complementary_filter",
    "ahrs_mahony_filter",
    "ahrs_hua_filter",
    "ahrs_madgwick_filter",
    "kf_lti_discretize",
    "kf_predict",
    "kf_update",
    "navdvl_dead_reckoning",
    "navdvl_kf",
    "UUV_EKF",
]
# math module
__all__ += [
    "filter_median",
    "filter_savgol",
    "filter_lowpass",
    "xyzrph2matrix",
    "matrix2xyzrph",
    "matrix_inverse",
    "pose_diff",
    "adjoint",
    "vec_to_se3",
    "se3_to_vec",
    "axis_ang6",
    "matrix_exp6",
    "matrix_log6",
    "project_to_se3",
    "distance_to_se3",
    "rph2rot",
    "hpr2rot",
    "rot2hpr",
    "rot2rph",
    "rot_diff",
    "vec_to_so3",
    "so3_to_vec",
    "so3_optimization",
    "axis_ang3",
    "matrix_exp3",
    "matrix_log3",
    "rot_inv",
    "project_to_so3",
    "distance_to_so3",
    "rph2quat",
    "hpr2quat",
    "quat2rph",
    "quat2hpr",
    "quat2rot",
    "rot2quat",
    "norm",
    "normalize",
    "mean",
    "median",
    "std",
    "min",
    "max",
    "print_stats",
    "remove_offset",
    "remove_mean",
    "remove_median",
    "difference",
    "derivative",
    "resample",
    "distance_traveled",
    "saturate",
    "wrap360",
    "wrap180",
    "wrap2pi",
    "wrap1pi",
    "unwrap2pi",
    "unwrap1pi",
    "unwrap360",
    "unwrap180",
    "wrapunwrap",
    "wrapunwrap360",
    "transpose",
    "right_qmatrix",
    "left_qmatrix",
    "quat_conj",
    "quat_inv",
    "quat_normalize",
    "quat_positive_scalar",
]
# lcmlog module
__all__ += [
    "msg_getfields",
    "msg_getconstants",
    "parse_and_save",
    "lcmlog_to_dict",
    "dict_to_struct",
    "read_log",
    "read_logs",
    "merge_lcm_logs",
    "filter_channels_from_data",
    "nav_structtype",
    "struct_to_dict",
]
# geo module
__all__ += [
    "latlon_to_zone_number",
    "latlon_to_zone_letter",
    "ll2utm",
    "utm2ll",
    "deg2m_lat",
    "deg2m_lon",
    "ll2xy",
    "get_local_gravity",
    "xy2ll",
    "dms_to_decimal",
]
# environment module
__all__ += [
    "ctd2svel",
    "depth2pressure",
    "ctp2salinity",
    "ctd2salinity",
    "ctp2pden",
    "ctd2oxygen_saturation",
]
# cal module
__all__ += [
    "cal_ahrs_so3",
    "cal_dvl_acc_so3",
    "cal_dvl_correct_by_sound_velocity",
    "cal_mag_ellipsoid_fit",
    "cal_mag_ellipsoid_fit_fang",
    "cal_mag_magfactor3",
    "cal_mag_twostep_hi",
    "cal_mag_twostep_hsi",
    "cal_mag_sphere_fit",
    "cal_mag_sar_ls",
    "cal_mag_sar_kf",
    "cal_mag_sar_aid",
    "cal_mag_magyc_ls",
    "cal_mag_magyc_nls",
    "cal_mag_magyc_bfg",
    "cal_mag_magyc_ifg",
    "cal_mag_magyc_ifg",
    "cal_dvl_jck_so3",
]
# mbsystem module
__all__ += [
    "mbs_read_fnv",
]

# logging module
__all__ += ["setup_logging"]



"""
Test module for the DVL calibration module.

Authors: Sebastián Rodríguez-Martínez
Contact: srodriguez@mbari.org
Last Update: 2025-04-15
"""

import numpy as np
import pytest

from navlib import cal_dvl_acc_so3, cal_dvl_correct_by_sound_velocity, cal_dvl_jck_so3
from navlib.cal import cal_dvl_acc_so3 as cal_dvl_acc_so3_relative
from navlib.cal import (
    cal_dvl_correct_by_sound_velocity as cal_dvl_correct_by_sound_velocity_relative,
)
from navlib.cal import cal_dvl_jck_so3 as cal_dvl_jck_so3_relative


def test_import():
    # Test if both import methods work
    assert cal_dvl_acc_so3_relative.__code__.co_code == cal_dvl_acc_so3.__code__.co_code
    assert (
        cal_dvl_correct_by_sound_velocity_relative.__code__.co_code
        == cal_dvl_correct_by_sound_velocity.__code__.co_code
    )
    assert cal_dvl_jck_so3_relative.__code__.co_code == cal_dvl_jck_so3.__code__.co_code


def dummy_inputs():
    # Create valid dummy inputs
    t = np.linspace(0, 1, 10)
    valid_time_dvl = t.copy()
    valid_time_ahrs = t.copy()
    valid_time_depth = t.copy()
    valid_velocity_dvl = np.ones((10, 3))
    valid_acceleration_ahrs = np.ones((10, 3))
    valid_rph_ahrs = np.ones((10, 3))
    valid_angular_rate_ahrs = np.ones((10, 3))
    valid_depth = t.copy()
    valid_lever_arm = np.array([0, 0, 0])

    return {
        "time_dvl": valid_time_dvl,
        "time_ahrs": valid_time_ahrs,
        "time_depth": valid_time_depth,
        "velocity_dvl": valid_velocity_dvl,
        "acceleration_ahrs": valid_acceleration_ahrs,
        "rph_ahrs": valid_rph_ahrs,
        "angular_rate_ahrs": valid_angular_rate_ahrs,
        "depth": valid_depth,
        "low_pass_filter": True,
        "remove_gravity": True,
        "lever_arm": valid_lever_arm,
        "method": "ls",
        "force_so3": True,
    }


def test_invalid_time_ahrs_type():
    params = dummy_inputs()
    params["time_ahrs"] = "not a numpy array"
    with pytest.raises(TypeError):
        cal_dvl_acc_so3(**params)


def test_invalid_low_pass_filter_type():
    params = dummy_inputs()
    params["low_pass_filter"] = "True"
    with pytest.raises(TypeError):
        cal_dvl_acc_so3(**params)


def test_invalid_time_ahrs_dimensions():
    params = dummy_inputs()
    # Provide a wrong dimension: 2-dimensional array that is not (n,) or (n,1) or (1,n)
    params["time_ahrs"] = np.ones((5, 5))
    with pytest.raises(ValueError):
        cal_dvl_acc_so3(**params)


def test_velocity_dvl_wrong_shape():
    params = dummy_inputs()
    # Provide velocity_dvl with wrong shape; should be 10x3 or 3x10 but here use 10x2.
    params["velocity_dvl"] = np.ones((10, 2))
    with pytest.raises(ValueError):
        cal_dvl_acc_so3(**params)


def test_mismatched_time_and_rph():
    params = dummy_inputs()
    # Make time_ahrs length different from rph_ahrs rows.
    params["time_ahrs"] = np.linspace(0, 1, 8)
    with pytest.raises(ValueError):
        cal_dvl_acc_so3(**params)


def test_mismatched_time_and_velocity():
    params = dummy_inputs()
    # Make time_dvl length different from velocity_dvl rows.
    params["time_dvl"] = np.linspace(0, 1, 12)
    with pytest.raises(ValueError):
        cal_dvl_acc_so3(**params)


def test_invalid_lever_arm_shape():
    params = dummy_inputs()
    # Provide lever_arm with wrong dimension; must have 3 elements.
    params["lever_arm"] = np.array([0, 0])
    with pytest.raises(ValueError):
        cal_dvl_acc_so3(**params)


def get_valid_inputs():
    # velocity_dvl: valid 10x3 array
    velocity_dvl = np.ones((10, 3))
    # conductivity, temperature, depth, latitude, longitude: matching 1D arrays of length 10
    conductivity = np.ones(10)
    temperature = np.ones(10) * 10  # arbitrary valid values
    depth = -100 * np.ones(10)  # negative depth values are required
    latitude = np.zeros(10)
    longitude = np.zeros(10)
    return velocity_dvl, conductivity, temperature, depth, latitude, longitude


def test_invalid_velocity_dvl_type():
    # velocity_dvl must be a list or numpy array; passing an int should trigger TypeError.
    (
        velocity_dvl,
        conductivity,
        temperature,
        depth,
        latitude,
        longitude,
    ) = get_valid_inputs()
    with pytest.raises(TypeError):
        cal_dvl_correct_by_sound_velocity(
            123, conductivity, temperature, depth, latitude, longitude
        )


def test_invalid_conductivity_type():
    (
        velocity_dvl,
        conductivity,
        temperature,
        depth,
        latitude,
        longitude,
    ) = get_valid_inputs()
    with pytest.raises(TypeError):
        cal_dvl_correct_by_sound_velocity(
            velocity_dvl, "not an array", temperature, depth, latitude, longitude
        )


def test_invalid_temperature_type():
    (
        velocity_dvl,
        conductivity,
        temperature,
        depth,
        latitude,
        longitude,
    ) = get_valid_inputs()
    with pytest.raises(TypeError):
        cal_dvl_correct_by_sound_velocity(
            velocity_dvl, conductivity, {"temp": 10}, depth, latitude, longitude
        )


def test_invalid_depth_type():
    (
        velocity_dvl,
        conductivity,
        temperature,
        depth,
        latitude,
        longitude,
    ) = get_valid_inputs()
    with pytest.raises(TypeError):
        cal_dvl_correct_by_sound_velocity(
            velocity_dvl, conductivity, temperature, "depth", latitude, longitude
        )


def test_invalid_latitude_type():
    (
        velocity_dvl,
        conductivity,
        temperature,
        depth,
        latitude,
        longitude,
    ) = get_valid_inputs()
    # latitude must be a numpy array or a float; passing a dict will raise.
    with pytest.raises(TypeError):
        cal_dvl_correct_by_sound_velocity(
            velocity_dvl,
            conductivity,
            temperature,
            depth,
            latitude={"lat": 0},
            longitude=longitude,
        )


def test_invalid_longitude_type():
    (
        velocity_dvl,
        conductivity,
        temperature,
        depth,
        latitude,
        longitude,
    ) = get_valid_inputs()
    with pytest.raises(TypeError):
        cal_dvl_correct_by_sound_velocity(
            velocity_dvl,
            conductivity,
            temperature,
            depth,
            latitude,
            longitude={"lon": 0},
        )


def test_velocity_shape_error():
    # velocity_dvl must be 2D and have one dimension size equal to 3.
    (
        velocity_dvl,
        conductivity,
        temperature,
        depth,
        latitude,
        longitude,
    ) = get_valid_inputs()
    bad_velocity = np.ones((5, 5))  # Both dims are not 3.
    with pytest.raises(ValueError):
        cal_dvl_correct_by_sound_velocity(
            bad_velocity, conductivity, temperature, depth, latitude, longitude
        )


def test_shape_mismatch_error():
    # Mismatch in number of measurements between velocity_dvl and other parameters.
    (
        velocity_dvl,
        conductivity,
        temperature,
        depth,
        latitude,
        longitude,
    ) = get_valid_inputs()
    bad_conductivity = np.ones(5)  # Length does not match velocity_dvl (10)
    with pytest.raises(ValueError):
        cal_dvl_correct_by_sound_velocity(
            velocity_dvl, bad_conductivity, temperature, depth, latitude, longitude
        )


def test_positive_depth_error():
    # Depth must be negative.
    (
        velocity_dvl,
        conductivity,
        temperature,
        depth,
        latitude,
        longitude,
    ) = get_valid_inputs()
    bad_depth = np.abs(depth)  # Make them positive
    with pytest.raises(ValueError):
        cal_dvl_correct_by_sound_velocity(
            velocity_dvl, conductivity, temperature, bad_depth, latitude, longitude
        )


def test_invalid_latitude_range():
    (
        velocity_dvl,
        conductivity,
        temperature,
        depth,
        latitude,
        longitude,
    ) = get_valid_inputs()
    # Set latitude out of valid range (< -90 or > 90)
    bad_latitude = -100 * np.ones_like(latitude)
    with pytest.raises(ValueError):
        cal_dvl_correct_by_sound_velocity(
            velocity_dvl, conductivity, temperature, depth, bad_latitude, longitude
        )


def test_invalid_longitude_range():
    (
        velocity_dvl,
        conductivity,
        temperature,
        depth,
        latitude,
        longitude,
    ) = get_valid_inputs()
    bad_longitude = 400 * np.ones_like(longitude)  # Out of valid range (-360,360)
    with pytest.raises(ValueError):
        cal_dvl_correct_by_sound_velocity(
            velocity_dvl, conductivity, temperature, depth, latitude, bad_longitude
        )


def get_valid_inputs_jck():
    """
    Generate a valid set of inputs for cal_dvl_jck.
    All time signals are 1D arrays of equal length. The velocity, rph, angular_rate
    and position arrays are shaped (n,3).
    """
    n = 10
    time_dvl = np.linspace(0, 9, n)
    time_ahrs = np.linspace(0, 9, n)
    time_position = np.linspace(0, 9, n)
    velocity_dvl = np.ones((n, 3))
    rph_ahrs = np.zeros((n, 3))
    angular_rate_ahrs = np.zeros((n, 3))
    position = 10 * np.ones((n, 3))
    return (
        time_dvl,
        time_ahrs,
        time_position,
        velocity_dvl,
        rph_ahrs,
        angular_rate_ahrs,
        position,
    )


def test_valid_ls_solution():
    """Test that a valid input returns a (3,3) matrix (LS method)."""
    inputs = get_valid_inputs_jck()
    R = cal_dvl_jck_so3(
        *inputs,
        low_pass_filter=False,
        method="ls",
        force_so3=False,
        zero_mean_sensor=False,
        zero_mean_equation=False,
    )
    assert R.shape == (3, 3)


def test_valid_svd_solution():
    """Test that a valid input returns a (3,3) matrix (SVD method)."""
    inputs = get_valid_inputs_jck()
    R = cal_dvl_jck_so3(
        *inputs,
        low_pass_filter=False,
        method="svd",
        force_so3=False,
        zero_mean_sensor=False,
        zero_mean_equation=False,
    )
    assert R.shape == (3, 3)


def test_invalid_time_dvl_type():
    """Test that a wrong type for time_dvl raises a TypeError."""
    inputs = list(get_valid_inputs_jck())
    inputs[0] = "not an array"  # time_dvl must be an array
    with pytest.raises(TypeError):
        cal_dvl_jck_so3(*inputs)


def test_invalid_velocity_shape():
    """Test that an invalid shape for velocity_dvl raises a ValueError."""
    inputs = list(get_valid_inputs_jck())
    # Force an incorrect shape (e.g. 5x5 instead of 3xN or Nx3)
    inputs[3] = np.ones((5, 5))
    with pytest.raises(ValueError):
        cal_dvl_jck_so3(*inputs)


def test_invalid_rph_ahrs_type():
    """Test that a wrong type for rph_ahrs raises a TypeError."""
    inputs = list(get_valid_inputs_jck())
    inputs[4] = "invalid"
    with pytest.raises(TypeError):
        cal_dvl_jck_so3(*inputs)


def test_invalid_position_shape():
    """Test that an invalid shape for position raises a ValueError."""
    inputs = list(get_valid_inputs_jck())
    # Provide a wrong shape (e.g. 10x10 instead of (n,3))
    inputs[6] = np.ones((10, 10))
    with pytest.raises(ValueError):
        cal_dvl_jck_so3(*inputs)



# import methods from the corresponding modules
from .filters import filter_lowpass, filter_median, filter_savgol
from .quaternion import (
    left_qmatrix,
    quat_conj,
    quat_inv,
    quat_normalize,
    quat_positive_scalar,
    right_qmatrix,
)
from .se3 import (
    adjoint,
    axis_ang6,
    distance_to_se3,
    matrix2xyzrph,
    matrix_exp6,
    matrix_inverse,
    matrix_log6,
    pose_diff,
    project_to_se3,
    se3_to_vec,
    vec_to_se3,
    xyzrph2matrix,
)
from .so3 import (
    axis_ang3,
    distance_to_so3,
    hpr2quat,
    hpr2rot,
    matrix_exp3,
    matrix_log3,
    project_to_so3,
    quat2hpr,
    quat2rot,
    quat2rph,
    rot2hpr,
    rot2quat,
    rot2rph,
    rot_diff,
    rot_inv,
    rph2quat,
    rph2rot,
    so3_optimization,
    so3_to_vec,
    vec_to_so3,
)
from .vmath import (
    derivative,
    difference,
    distance_traveled,
    max,
    mean,
    median,
    min,
    norm,
    normalize,
    print_stats,
    remove_mean,
    remove_median,
    remove_offset,
    resample,
    saturate,
    std,
    transpose,
    unwrap1pi,
    unwrap2pi,
    unwrap180,
    unwrap360,
    wrap1pi,
    wrap2pi,
    wrap180,
    wrap360,
    wrapunwrap,
    wrapunwrap360,
)

# Get __all__ from the corresponding modules
# filters module
__all__ = [
    "filter_median",
    "filter_savgol",
    "filter_lowpass",
]
# quaternion module
__all__ += [
    "left_qmatrix",
    "right_qmatrix",
    "quat_conj",
    "quat_inv",
    "quat_normalize",
    "quat_positive_scalar",
]
# se3 module
__all__ += [
    "xyzrph2matrix",
    "matrix2xyzrph",
    "matrix_inverse",
    "pose_diff",
    "adjoint",
    "vec_to_se3",
    "se3_to_vec",
    "axis_ang6",
    "matrix_exp6",
    "matrix_log6",
    "project_to_se3",
    "distance_to_se3",
]
# so3 module
__all__ += [
    "rph2rot",
    "hpr2rot",
    "rot2hpr",
    "rot2rph",
    "rot_diff",
    "vec_to_so3",
    "so3_to_vec",
    "axis_ang3",
    "matrix_exp3",
    "matrix_log3",
    "rot_inv",
    "project_to_so3",
    "distance_to_so3",
    "rph2quat",
    "hpr2quat",
    "quat2rph",
    "quat2hpr",
    "quat2rot",
    "rot2quat",
    "so3_optimization",
]
# vmath module
__all__ += [
    "norm",
    "normalize",
    "mean",
    "median",
    "std",
    "min",
    "max",
    "print_stats",
    "remove_offset",
    "remove_mean",
    "remove_median",
    "difference",
    "derivative",
    "resample",
    "distance_traveled",
    "saturate",
    "wrap360",
    "wrap180",
    "wrap2pi",
    "wrap1pi",
    "unwrap2pi",
    "unwrap1pi",
    "unwrap360",
    "unwrap180",
    "wrapunwrap",
    "wrapunwrap360",
    "transpose",
]
